'''Retrieve Air Quality Forecasts'''
import warnings
from datetime import date as date_, datetime
from typing import Callable, Coroutine, Optional, Union

from pyairnow.util import coalesce


PARAMETER_NAME_MAP = {
    'OZONE': 'O3',
}


def _normalize_forecast(raw: dict) -> dict:
    '''Transform new API response format to legacy format for compatibility.'''
    legacy_cat = coalesce(raw, 'Category', default={})
    cat_number = coalesce(raw, 'categoryNumber')
    if cat_number is None:
        cat_number = coalesce(legacy_cat, 'Number', default=0)
    cat_name = coalesce(raw, 'categoryName')
    if cat_name is None:
        cat_name = coalesce(legacy_cat, 'Name', default='')

    param = coalesce(raw, 'parameterName', 'ParameterName', default='')
    param = PARAMETER_NAME_MAP.get(param, param)

    return {
        'DateIssue': coalesce(raw, 'dateIssue', default=''),
        'DateForecast': coalesce(raw, 'dateValid', 'DateForecast', default=''),
        'ReportingArea': coalesce(
            raw, 'reportingArea', 'ReportingArea', default=''
        ),
        'StateCode': coalesce(raw, 'stateCode', 'StateCode', default=''),
        'Latitude': coalesce(raw, 'latitude', 'Latitude'),
        'Longitude': coalesce(raw, 'longitude', 'Longitude'),
        'ParameterName': param,
        'AQI': coalesce(raw, 'aqi', 'AQI', default=-1),
        'Category': {
            'Number': cat_number,
            'Name': cat_name,
        },
        'ActionDay': coalesce(raw, 'actionDay', 'ActionDay', default=False),
        'Discussion': coalesce(raw, 'discussion', 'Discussion', default=''),
    }


class Forecast:
    '''
    Class to retrieve the air quality forecast by zip code or by latitude and
    longitude.
    '''
    def __init__(
        self, request: Callable[..., Coroutine], *,
        legacy_format: bool = True
    ) -> None:
        self._request = request
        self._legacy_format = legacy_format

    async def zipCode(
        self,
        zipCode: str,
        *,
        date: Optional[Union[date_, datetime, str]] = None,
        distance: Optional[int] = None
    ) -> list:
        '''Request forecast for zip code'''
        params: dict = dict(zipCode=zipCode)
        if date and isinstance(date, str):
            y, m, d = date.split('-')
            params['date'] = date_(int(y), int(m), int(d)).isoformat()
        elif date and isinstance(date, datetime):
            params['date'] = date.date().isoformat()
        elif date and isinstance(date, date_):
            params['date'] = date.isoformat()
        if distance is not None:
            warnings.warn(
                'The distance parameter is deprecated and ignored by the '
                'AirNow 2026 API. It will be removed in a future version.',
                DeprecationWarning,
                stacklevel=2,
            )

        data = await self._request(
            'aq/forecast/current',
            params=params
        )
        if self._legacy_format:
            return [_normalize_forecast(f) for f in data]
        return data

    async def latLong(
        self,
        latitude: Optional[Union[float, str]] = None,
        longitude: Optional[Union[float, str]] = None,
        *,
        date: Optional[Union[date_, datetime, str]] = None,
        distance: Optional[int] = None,
    ) -> list:
        '''Request forecast for latitude/longitude'''
        params: dict = dict(
            latitude=str(latitude),
            longitude=str(longitude),
        )
        if date and isinstance(date, str):
            y, m, d = date.split('-')
            params['date'] = date_(int(y), int(m), int(d)).isoformat()
        elif date and isinstance(date, datetime):
            params['date'] = date.date().isoformat()
        elif date and isinstance(date, date_):
            params['date'] = date.isoformat()
        if distance is not None:
            warnings.warn(
                'The distance parameter is deprecated and ignored by the '
                'AirNow 2026 API. It will be removed in a future version.',
                DeprecationWarning,
                stacklevel=2,
            )

        data = await self._request(
            'aq/forecast/current',
            params=params
        )
        if self._legacy_format:
            return [_normalize_forecast(f) for f in data]
        return data



import datetime
import pytest

from pyairnow import WebServiceAPI

from .mock_api import MOCK_API_KEY


@pytest.mark.asyncio
async def test_api_forecast_zipcode(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 2

    assert data[0]['DateIssue'] == '2020-09-01'
    assert data[0]['DateForecast'] == '2020-09-02'
    assert data[0]['ReportingArea'] == 'Los Angeles'
    assert data[0]['StateCode'] == 'CA'
    assert data[0]['ParameterName'] == 'O3'
    assert data[0]['AQI'] == 55
    assert data[0]['Category'] == {'Number': 2, 'Name': 'Moderate'}
    assert data[0]['ActionDay'] is False

    assert data[1]['ParameterName'] == 'PM2.5'
    assert data[1]['AQI'] == 42
    assert data[1]['Category'] == {'Number': 1, 'Name': 'Good'}


@pytest.mark.asyncio
async def test_api_forecast_zipcode_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.warns(DeprecationWarning, match='distance parameter is deprecated'):
        data = await client.forecast.zipCode(90001, distance=100)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['ParameterName'] == 'O3'


@pytest.mark.asyncio
async def test_api_forecast_zipcode_date_str(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(90001, date='2020-09-01')

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_zipcode_date_date(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(
        90001,
        date=datetime.date(2020, 9, 1)
    )

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_zipcode_date_datetime(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(
        90001,
        date=datetime.datetime(2020, 9, 1, 11, 45, 1)
    )

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_ll(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(34.053718, -118.244842)

    assert isinstance(data, list)
    assert len(data) == 2

    assert data[0]['DateIssue'] == '2020-09-01'
    assert data[0]['DateForecast'] == '2020-09-02'
    assert data[0]['ReportingArea'] == 'Los Angeles'
    assert data[0]['StateCode'] == 'CA'
    assert data[0]['ParameterName'] == 'O3'
    assert data[0]['AQI'] == 55
    assert data[0]['Category'] == {'Number': 2, 'Name': 'Moderate'}


@pytest.mark.asyncio
async def test_api_forecast_ll_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.warns(DeprecationWarning, match='distance parameter is deprecated'):
        data = await client.forecast.latLong(
            34.053718, -118.244842,
            distance=120
        )

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_ll_date_str(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        date='2020-09-01'
    )

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_ll_date_date(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        date=datetime.date(2020, 9, 1)
    )

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_ll_date_datetime(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        date=datetime.datetime(2020, 9, 1, 11, 45, 0)
    )

    assert isinstance(data, list)
    assert len(data) == 2


@pytest.mark.asyncio
async def test_api_forecast_zipcode_raw_format(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY, legacy_format=False)
    data = await client.forecast.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['parameterName'] == 'OZONE'
    assert data[0]['aqi'] == 55
    assert data[0]['categoryName'] == 'Moderate'
    assert 'DateIssue' not in data[0]


@pytest.mark.asyncio
async def test_api_forecast_ll_raw_format(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY, legacy_format=False)
    data = await client.forecast.latLong(34.053718, -118.244842)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['parameterName'] == 'OZONE'
    assert data[0]['aqi'] == 55
    assert 'DateIssue' not in data[0]



'''Mocked AirNow API'''
import re

from aioresponses import CallbackResult

MOCK_API_KEY = '01234567-89AB-CDEF-0123-456789ABCDEF'

RE_FORECAST_CURRENT = re.compile(r'^/?aq/forecast/current/?$')
RE_OBS_ZIPLATLONG = re.compile(r'^/?aq/observation/current/ziplatlong/?$')

MOCK_OBSERVATION_RESPONSE = [
    {
        'dateObserved': '2020-09-01',
        'hourObserved': '15:00',
        'localTimeZone': 'PDT',
        'reportingAreaName': 'Los Angeles',
        'siteID': '060371103',
        'siteName': 'Los Angeles - N Main',
        'parameterName': 'PM2.5',
        'nowcastAQI': 55,
        'aqiCategoryName': 'Moderate',
        'reportingAgency': 'South Coast AQMD',
        'lookupBehavior': 'Closest Reading By Pollutant',
        'consideredMonitors': 'All',
        'lookupBoundary': '25 Miles',
    },
    {
        'dateObserved': '2020-09-01',
        'hourObserved': '15:00',
        'localTimeZone': 'PDT',
        'reportingAreaName': 'Los Angeles',
        'siteID': '060371103',
        'siteName': 'Los Angeles - N Main',
        'parameterName': 'OZONE',
        'nowcastAQI': 42,
        'aqiCategoryName': 'Good',
        'reportingAgency': 'South Coast AQMD',
        'lookupBehavior': 'Closest Reading By Pollutant',
        'consideredMonitors': 'All',
        'lookupBoundary': '25 Miles',
    },
]

MOCK_FORECAST_RESPONSE = [
    {
        'dateIssue': '2020-09-01',
        'dateValid': '2020-09-02',
        'reportingArea': 'Los Angeles',
        'reportingAreaCode': 'ca060',
        'stateCode': 'CA',
        'parameterName': 'OZONE',
        'aqi': 55,
        'forecastAgency': 'South Coast AQMD',
        'categoryNumber': 2,
        'categoryName': 'Moderate',
        'actionDay': False,
        'discussion': '',
    },
    {
        'dateIssue': '2020-09-01',
        'dateValid': '2020-09-02',
        'reportingArea': 'Los Angeles',
        'reportingAreaCode': 'ca060',
        'stateCode': 'CA',
        'parameterName': 'PM2.5',
        'aqi': 42,
        'forecastAgency': 'South Coast AQMD',
        'categoryNumber': 1,
        'categoryName': 'Good',
        'actionDay': False,
        'discussion': '',
    },
]


def mock_airnow_api(url, **kwargs):
    '''
    Mock the AirNow API
    '''
    # Check API Key
    if 'API_KEY' not in url.query:
        return CallbackResult(status=401, payload=dict(
            WebServiceError=[
                dict(
                    Message='Request not authenticated'
                ),
            ],
        ))

    if url.query['API_KEY'] != MOCK_API_KEY:
        return CallbackResult(status=401, payload=dict(
            WebServiceError=[
                dict(
                    Message='Invalid API key'
                ),
            ],
        ))

    # A zip code with value "empty" will return an empty list
    if 'zipCode' in url.query and url.query['zipCode'] == 'empty':
        return CallbackResult(payload=[])

    # A zip code with value "bad_json" will return invalid json
    if 'zipCode' in url.query and url.query['zipCode'] == 'bad_json':
        return CallbackResult(body='Bad JSON Test')

    # A zip code with value "error" will return a JSON error message
    if 'zipCode' in url.query and url.query['zipCode'] == 'error':
        return CallbackResult(status=400, payload=dict(
            WebServiceError=[
                dict(
                    Message='Client Error'
                ),
            ],
        ))

    # A zip code with value "error1" will return a JSON error message
    if 'zipCode' in url.query and url.query['zipCode'] == 'error1':
        return CallbackResult(status=400, payload=dict(
            WebServiceError=[
                'Client Error'
            ],
        ))

    # A zip code with value "error2" will return a non-structured error message
    if 'zipCode' in url.query and url.query['zipCode'] == 'error2':
        return CallbackResult(status=500, payload=dict(
            WebServiceError='Internal Server Error'
        ))

    # A zip code with value "dict" will return a JSON dictionary instead of
    # an array
    if 'zipCode' in url.query and url.query['zipCode'] == 'dict':
        return CallbackResult(payload=dict(status='OK'))

    # Route to appropriate mock response
    if RE_OBS_ZIPLATLONG.match(url.path):
        return CallbackResult(payload=MOCK_OBSERVATION_RESPONSE)

    if RE_FORECAST_CURRENT.match(url.path):
        return CallbackResult(payload=MOCK_FORECAST_RESPONSE)

    return CallbackResult(status=404)



'''Tests for shared normalization helpers'''
from pyairnow.util import coalesce


class TestCoalesce:
    def test_returns_first_present_value(self):
        assert coalesce({'a': 1, 'b': 2}, 'a', 'b') == 1

    def test_skips_null_value(self):
        assert coalesce({'a': None, 'b': 2}, 'a', 'b') == 2

    def test_skips_absent_key(self):
        assert coalesce({'b': 2}, 'a', 'b') == 2

    def test_default_when_all_null(self):
        assert coalesce({'a': None, 'b': None}, 'a', 'b', default='x') == 'x'

    def test_default_when_all_absent(self):
        assert coalesce({}, 'a', 'b', default='x') == 'x'

    def test_default_is_none_when_unset(self):
        assert coalesce({}, 'a') is None

    def test_falsy_values_are_not_skipped(self):
        assert coalesce({'a': 0}, 'a', default=-1) == 0
        assert coalesce({'a': False}, 'a', default=True) is False
        assert coalesce({'a': ''}, 'a', default='x') == ''



'''Retrieve a list of Current Observations'''
import warnings
from typing import Callable, Coroutine, Optional, Union

from pyairnow.util import coalesce


CATEGORY_NAME_TO_NUMBER = {
    'Good': 1,
    'Moderate': 2,
    'Unhealthy for Sensitive Groups': 3,
    'Unhealthy': 4,
    'Very Unhealthy': 5,
    'Hazardous': 6,
}


PARAMETER_NAME_MAP = {
    'OZONE': 'O3',
}


def _normalize_observation(raw: dict) -> dict:
    '''Transform new API response format to legacy format for compatibility.'''
    hour_raw = coalesce(raw, 'hourObserved', default=0)
    if isinstance(hour_raw, str):
        hour = int(hour_raw.split(':')[0])
    else:
        hour = int(hour_raw)

    cat_name = coalesce(raw, 'aqiCategoryName', default='')
    cat_number = CATEGORY_NAME_TO_NUMBER.get(cat_name, 0)

    param = coalesce(raw, 'parameterName', default='')
    param = PARAMETER_NAME_MAP.get(param, param)

    return {
        'DateObserved': coalesce(raw, 'dateObserved', default=''),
        'HourObserved': hour,
        'LocalTimeZone': coalesce(raw, 'localTimeZone', default=''),
        'ReportingArea': coalesce(raw, 'reportingAreaName', default=''),
        'StateCode': coalesce(raw, 'stateCode', default=''),
        'Latitude': coalesce(raw, 'latitude'),
        'Longitude': coalesce(raw, 'longitude'),
        'ParameterName': param,
        'AQI': coalesce(raw, 'nowcastAQI', 'aqi', default=-1),
        'Category': {
            'Number': cat_number,
            'Name': cat_name,
        },
    }


class Observations:
    '''
    Class to retrieve the current air quality observations by zip code or by
    latitude and longitude.
    '''
    def __init__(
        self, request: Callable[..., Coroutine], *,
        legacy_format: bool = True
    ) -> None:
        self._request = request
        self._legacy_format = legacy_format

    async def zipCode(
        self,
        zipCode: str,
        *,
        distance: Optional[int] = None
    ) -> list:
        '''Request current observation for zip code'''
        params: dict = dict(zipCode=zipCode)
        if distance is not None:
            warnings.warn(
                'The distance parameter is deprecated and ignored by the '
                'AirNow 2026 API. It will be removed in a future version.',
                DeprecationWarning,
                stacklevel=2,
            )

        data = await self._request(
            'aq/observation/current/ziplatlong',
            params=params
        )
        if self._legacy_format:
            return [_normalize_observation(o) for o in data]
        return data

    async def latLong(
        self,
        latitude: Optional[Union[float, str]] = None,
        longitude: Optional[Union[float, str]] = None,
        *,
        distance: Optional[int] = None,
    ) -> list:
        '''Request current observation for latitude/longitude'''
        params: dict = dict(
            latitude=str(latitude),
            longitude=str(longitude),
        )
        if distance is not None:
            warnings.warn(
                'The distance parameter is deprecated and ignored by the '
                'AirNow 2026 API. It will be removed in a future version.',
                DeprecationWarning,
                stacklevel=2,
            )

        data = await self._request(
            'aq/observation/current/ziplatlong',
            params=params
        )
        if self._legacy_format:
            return [_normalize_observation(o) for o in data]
        return data



'''Tests for response normalization from new API format to legacy format'''
from pyairnow.observation import _normalize_observation, CATEGORY_NAME_TO_NUMBER
from pyairnow.forecast import _normalize_forecast


class TestObservationNormalization:
    def test_basic_fields(self):
        raw = {
            'dateObserved': '2024-01-15',
            'hourObserved': '14:00',
            'localTimeZone': 'PST',
            'reportingAreaName': 'San Francisco',
            'parameterName': 'PM2.5',
            'nowcastAQI': 42,
            'aqiCategoryName': 'Good',
        }
        result = _normalize_observation(raw)

        assert result['DateObserved'] == '2024-01-15'
        assert result['HourObserved'] == 14
        assert result['LocalTimeZone'] == 'PST'
        assert result['ReportingArea'] == 'San Francisco'
        assert result['ParameterName'] == 'PM2.5'
        assert result['AQI'] == 42
        assert result['Category'] == {'Number': 1, 'Name': 'Good'}

    def test_ozone_renamed_to_o3(self):
        raw = {
            'parameterName': 'OZONE',
            'nowcastAQI': 30,
            'aqiCategoryName': 'Good',
        }
        result = _normalize_observation(raw)
        assert result['ParameterName'] == 'O3'

    def test_pm10_unchanged(self):
        raw = {'parameterName': 'PM10', 'nowcastAQI': 20, 'aqiCategoryName': 'Good'}
        result = _normalize_observation(raw)
        assert result['ParameterName'] == 'PM10'

    def test_hour_string_parsed(self):
        raw = {'hourObserved': '09:00', 'aqiCategoryName': ''}
        result = _normalize_observation(raw)
        assert result['HourObserved'] == 9

    def test_hour_integer_passthrough(self):
        raw = {'hourObserved': 15, 'aqiCategoryName': ''}
        result = _normalize_observation(raw)
        assert result['HourObserved'] == 15

    def test_all_category_mappings(self):
        for name, expected_number in CATEGORY_NAME_TO_NUMBER.items():
            raw = {'aqiCategoryName': name}
            result = _normalize_observation(raw)
            assert result['Category']['Number'] == expected_number
            assert result['Category']['Name'] == name

    def test_unknown_category_defaults_to_zero(self):
        raw = {'aqiCategoryName': 'Unknown Category'}
        result = _normalize_observation(raw)
        assert result['Category']['Number'] == 0

    def test_missing_fields_default_gracefully(self):
        result = _normalize_observation({})
        assert result['DateObserved'] == ''
        assert result['HourObserved'] == 0
        assert result['LocalTimeZone'] == ''
        assert result['ReportingArea'] == ''
        assert result['StateCode'] == ''
        assert result['Latitude'] is None
        assert result['Longitude'] is None
        assert result['ParameterName'] == ''
        assert result['AQI'] == -1

    def test_null_fields_default_gracefully(self):
        raw = {
            'dateObserved': None,
            'hourObserved': None,
            'localTimeZone': None,
            'reportingAreaName': None,
            'stateCode': None,
            'latitude': None,
            'longitude': None,
            'parameterName': None,
            'nowcastAQI': None,
            'aqiCategoryName': None,
        }
        result = _normalize_observation(raw)
        assert result['DateObserved'] == ''
        assert result['HourObserved'] == 0
        assert result['LocalTimeZone'] == ''
        assert result['ReportingArea'] == ''
        assert result['StateCode'] == ''
        assert result['Latitude'] is None
        assert result['Longitude'] is None
        assert result['ParameterName'] == ''
        assert result['AQI'] == -1
        assert result['Category'] == {'Number': 0, 'Name': ''}

    def test_null_aqi_falls_back_to_aqi_key(self):
        raw = {'nowcastAQI': None, 'aqi': 42, 'aqiCategoryName': 'Good'}
        result = _normalize_observation(raw)
        assert result['AQI'] == 42

    def test_zero_aqi_is_preserved(self):
        raw = {'nowcastAQI': 0, 'aqiCategoryName': 'Good'}
        result = _normalize_observation(raw)
        assert result['AQI'] == 0


class TestForecastNormalization:
    def test_basic_fields(self):
        raw = {
            'dateIssue': '2024-01-15',
            'dateValid': '2024-01-16',
            'reportingArea': 'Los Angeles',
            'stateCode': 'CA',
            'parameterName': 'PM2.5',
            'aqi': 55,
            'categoryNumber': 2,
            'categoryName': 'Moderate',
            'actionDay': True,
            'discussion': 'Expect moderate air quality',
        }
        result = _normalize_forecast(raw)

        assert result['DateIssue'] == '2024-01-15'
        assert result['DateForecast'] == '2024-01-16'
        assert result['ReportingArea'] == 'Los Angeles'
        assert result['StateCode'] == 'CA'
        assert result['ParameterName'] == 'PM2.5'
        assert result['AQI'] == 55
        assert result['Category'] == {'Number': 2, 'Name': 'Moderate'}
        assert result['ActionDay'] is True
        assert result['Discussion'] == 'Expect moderate air quality'

    def test_ozone_renamed_to_o3(self):
        raw = {
            'parameterName': 'OZONE',
            'aqi': 30,
            'categoryNumber': 1,
            'categoryName': 'Good',
        }
        result = _normalize_forecast(raw)
        assert result['ParameterName'] == 'O3'

    def test_missing_fields_default_gracefully(self):
        result = _normalize_forecast({})
        assert result['DateIssue'] == ''
        assert result['DateForecast'] == ''
        assert result['ReportingArea'] == ''
        assert result['StateCode'] == ''
        assert result['ParameterName'] == ''
        assert result['AQI'] == -1
        assert result['Category'] == {'Number': 0, 'Name': ''}
        assert result['ActionDay'] is False
        assert result['Discussion'] == ''

    def test_null_fields_default_gracefully(self):
        raw = {
            'dateIssue': None,
            'dateValid': None,
            'reportingArea': None,
            'stateCode': None,
            'latitude': None,
            'longitude': None,
            'parameterName': None,
            'aqi': None,
            'categoryNumber': None,
            'categoryName': None,
            'actionDay': None,
            'discussion': None,
        }
        result = _normalize_forecast(raw)
        assert result['DateIssue'] == ''
        assert result['DateForecast'] == ''
        assert result['ReportingArea'] == ''
        assert result['StateCode'] == ''
        assert result['Latitude'] is None
        assert result['Longitude'] is None
        assert result['ParameterName'] == ''
        assert result['AQI'] == -1
        assert result['Category'] == {'Number': 0, 'Name': ''}
        assert result['ActionDay'] is False
        assert result['Discussion'] == ''

    def test_null_legacy_category_defaults_gracefully(self):
        result = _normalize_forecast({'Category': None, 'AQI': 20})
        assert result['Category'] == {'Number': 0, 'Name': ''}
        assert result['AQI'] == 20

    def test_legacy_category_still_read(self):
        name = 'Unhealthy for Sensitive Groups'
        result = _normalize_forecast({'Category': {'Number': 3, 'Name': name}})
        assert result['Category']['Number'] == 3
        assert result['Category']['Name'] == name

    def test_action_day_false_is_preserved(self):
        result = _normalize_forecast({'actionDay': False})
        assert result['ActionDay'] is False



import pytest

from pyairnow import WebServiceAPI

from .mock_api import MOCK_API_KEY


@pytest.mark.asyncio
async def test_api_observations_zipcode(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.observations.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 2

    assert data[0]['DateObserved'] == '2020-09-01'
    assert data[0]['HourObserved'] == 15
    assert data[0]['LocalTimeZone'] == 'PDT'
    assert data[0]['ReportingArea'] == 'Los Angeles'
    assert data[0]['ParameterName'] == 'PM2.5'
    assert data[0]['AQI'] == 55
    assert data[0]['Category'] == {'Number': 2, 'Name': 'Moderate'}

    assert data[1]['ParameterName'] == 'O3'
    assert data[1]['AQI'] == 42
    assert data[1]['Category'] == {'Number': 1, 'Name': 'Good'}


@pytest.mark.asyncio
async def test_api_observations_zipcode_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.warns(DeprecationWarning, match='distance parameter is deprecated'):
        data = await client.observations.zipCode(90001, distance=100)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['ParameterName'] == 'PM2.5'


@pytest.mark.asyncio
async def test_api_observations_ll(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.observations.latLong(34.053718, -118.244842)

    assert isinstance(data, list)
    assert len(data) == 2

    assert data[0]['DateObserved'] == '2020-09-01'
    assert data[0]['HourObserved'] == 15
    assert data[0]['LocalTimeZone'] == 'PDT'
    assert data[0]['ReportingArea'] == 'Los Angeles'
    assert data[0]['ParameterName'] == 'PM2.5'
    assert data[0]['AQI'] == 55
    assert data[0]['Category'] == {'Number': 2, 'Name': 'Moderate'}


@pytest.mark.asyncio
async def test_api_observations_ll_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.warns(DeprecationWarning, match='distance parameter is deprecated'):
        data = await client.observations.latLong(
            34.053718, -118.244842,
            distance=120
        )

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['ParameterName'] == 'PM2.5'


@pytest.mark.asyncio
async def test_api_observations_zipcode_raw_format(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY, legacy_format=False)
    data = await client.observations.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['parameterName'] == 'PM2.5'
    assert data[0]['nowcastAQI'] == 55
    assert data[0]['aqiCategoryName'] == 'Moderate'
    assert 'DateObserved' not in data[0]


@pytest.mark.asyncio
async def test_api_observations_ll_raw_format(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY, legacy_format=False)
    data = await client.observations.latLong(34.053718, -118.244842)

    assert isinstance(data, list)
    assert len(data) == 2
    assert data[0]['parameterName'] == 'PM2.5'
    assert data[0]['nowcastAQI'] == 55
    assert 'DateObserved' not in data[0]



'''Shared helpers for response normalization'''
from typing import Any


def coalesce(raw: dict, *keys: str, default: Any = None) -> Any:
    '''
    Return the first non-null value among `keys`, or `default`.

    `dict.get(key, default)` only falls back when the key is absent, so a key
    present with a JSON null returns None rather than the default.
    '''
    for key in keys:
        value = raw.get(key)
        if value is not None:
            return value
    return default
